import asyncio
import sys

from .exceptions import LoopNotRunning


def _get_current_task(loop):  # noqa
    if sys.version_info[:2] < (3, 7):  # to avoid deprecation warning
        return asyncio.Task.current_task(loop=loop)

    else:
        return asyncio.current_task(loop=loop)


def _check_event_loop():
    if not asyncio.get_event_loop().is_running():
        raise LoopNotRunning


def _get_or_create_cache():
    loop = asyncio.get_event_loop()
    try:
        return _get_current_task(loop)._sanicjwt

    except AttributeError:
        _get_current_task(loop)._sanicjwt = {}
        return _get_current_task(loop)._sanicjwt


def get_cached(value):
    _check_event_loop()
    return _get_or_create_cache().get(value, None)


def is_cached(value):
    _check_event_loop()
    return value in _get_or_create_cache()


def to_cache(key, value):
    _check_event_loop()
    _get_or_create_cache().update({key: value})


def clear_cache():
    _check_event_loop()
    loop = asyncio.get_event_loop()
    try:
        _get_current_task(loop)._sanicjwt = {}
    except AttributeError:  # noqa
        pass



from sanic.response import json

from .base import BaseDerivative

COOKIE_OPTIONS = (
    ("domain", "cookie_domain"),
    ("expires", "cookie_expires"),
    ("max-age", "cookie_max_age"),
    ("secure", "cookie_secure"),
)


def _set_cookie(response, key, value, config, force_httponly=None):
    response.cookies[key] = value
    response.cookies[key]["httponly"] = (
        config.cookie_httponly() if force_httponly is None else force_httponly
    )
    response.cookies[key]["path"] = config.cookie_path()

    for item, option in COOKIE_OPTIONS:
        value = getattr(config, option)()
        if value:
            response.cookies[key][item] = value


class Responses(BaseDerivative):
    @staticmethod
    async def get_access_token_output(request, user, config, instance):
        access_token = await instance.auth.generate_access_token(user)

        output = {config.access_token_name(): access_token}

        return access_token, output

    @staticmethod
    def get_token_response(
        request, access_token, output, config, refresh_token=None
    ):
        response = json(output)

        if config.cookie_set():
            key = config.cookie_access_token_name()

            if config.cookie_split():
                signature_name = config.cookie_split_signature_name()
                header_payload, signature = access_token.rsplit(
                    ".", maxsplit=1
                )
                _set_cookie(
                    response, key, header_payload, config, force_httponly=False
                )
                _set_cookie(
                    response,
                    signature_name,
                    signature,
                    config,
                    force_httponly=True,
                )
            else:
                _set_cookie(response, key, access_token, config)

            if refresh_token and config.refresh_token_enabled():
                key = config.cookie_refresh_token_name()
                _set_cookie(response, key, refresh_token, config)

        return response

    @staticmethod
    def extend_authenticate(
        request, user=None, access_token=None, refresh_token=None
    ):
        return {}

    @staticmethod
    def extend_retrieve_user(request, user=None, payload=None):
        return {}

    @staticmethod
    def extend_verify(request, user=None, payload=None):
        return {}

    @staticmethod
    def extend_refresh(
        request,
        user=None,
        access_token=None,
        refresh_token=None,
        purported_token=None,
        payload=None,
    ):
        return {}

    @staticmethod
    def exception_response(request, exception):
        reasons = (
            exception.args[0]
            if isinstance(exception.args[0], list)
            else [exception.args[0]]
        )
        return json(
            {"reasons": reasons, "exception": exception.__class__.__name__},
            status=exception.status_code,
        )



import logging
from contextlib import contextmanager
from copy import deepcopy
from functools import wraps
from inspect import isawaitable

from sanic import Blueprint
from sanic.response import redirect
from sanic.views import HTTPMethodView

from . import exceptions, utils
from .cache import clear_cache, to_cache
from .validators import validate_scopes

logger = logging.getLogger(__name__)


@contextmanager
def instant_config(instance, **kwargs):
    if kwargs and hasattr(instance, "auth"):
        to_cache("_request", kwargs.get("request"))
        for key, val in kwargs.items():
            if key in instance.auth.config:
                if callable(val):
                    val = val()
                to_cache(key, val)
    yield

    clear_cache()


async def _do_protection(*args, **kwargs):
    initialized_on = kwargs.pop("initialized_on")
    kw = kwargs.pop("kw")
    request = kwargs.pop("request")
    f = kwargs.pop("f")

    use_kwargs = deepcopy(kwargs)
    if "return_response" in use_kwargs:
        use_kwargs.pop("return_response")

    if initialized_on and isinstance(initialized_on, Blueprint):
        instance = initialized_on
    else:
        instance = request.app

    with instant_config(instance, request=request, **kw):
        if request.method == "OPTIONS":
            response = f(request, *args, **use_kwargs)
            if isawaitable(response):  # noqa
                response = await response
            if kwargs.get("return_response", True):
                return response

            else:
                return True, response

        try:
            if instance.auth.config.do_protection():
                (
                    is_authenticated,
                    status,
                    reasons,
                ) = await instance.auth._check_authentication(
                    request, request_args=args, request_kwargs=use_kwargs
                )
            else:
                is_authenticated = True
                status = 200
                reasons = None
        except AttributeError:
            raise exceptions.SanicJWTException(
                "Authentication instance not found. Perhaps you used "
                "@protected without passing in a blueprint? "
                "Try @protected(blueprint)",
                status_code=500,
            )

        except exceptions.SanicJWTException as e:
            is_authenticated = False
            status = e.status_code
            reasons = (
                instance.auth._reasons
                if (instance.auth._reasons and instance.auth.config.debug())
                else e.args[0]
            )

        if is_authenticated:
            if kwargs.get("return_response", True):
                response = f(request, *args, **use_kwargs)
                if isawaitable(response):
                    response = await response
                return response

            else:
                return True, instance

        else:
            if kw.get("redirect_on_fail", False):
                where_to = kw.get(
                    "redirect_url", instance.auth.config.login_redirect_url()
                )

                if where_to is not None:
                    return redirect(where_to, status=302)

            raise exceptions.Unauthorized(reasons, status_code=status)


def protected(initialized_on=None, **kw):
    def decorator(f):
        @wraps(f)
        async def decorated_function(request, *args, **kwargs):
            if issubclass(request.__class__, HTTPMethodView):
                request = args[0]
            kwargs.update(
                {
                    "initialized_on": initialized_on,
                    "kw": kw,
                    "request": request,
                    "f": f,
                }
            )
            return await _do_protection(*args, **kwargs)

        return decorated_function

    return decorator


def scoped(
    scopes,
    require_all=True,
    require_all_actions=True,
    initialized_on=None,
    **kw
):
    def decorator(f):
        @wraps(f)
        async def decorated_function(request, *args, **kwargs):
            if issubclass(request.__class__, HTTPMethodView):
                request = args[0]

            if scopes is not None and scopes is not False:
                protect_kwargs = deepcopy(kwargs)
                protect_kwargs.update(
                    {
                        "initialized_on": initialized_on,
                        "kw": kw,
                        "request": request,
                        "f": f,
                        "return_response": False,
                    }
                )
                _, instance = await _do_protection(*args, **protect_kwargs)

                if request.method == "OPTIONS":
                    return instance

                with instant_config(instance, request=request, **kw):
                    user_scopes = await instance.auth.extract_scopes(request)
                    override = instance.auth.override_scope_validator
                    destructure = instance.auth.destructure_scopes
                    if user_scopes is None:
                        # If there are no defined scopes in the payload,
                        # deny access
                        is_authorized = False
                        status = 403
                        reasons = "Invalid scope."

                        # TODO:
                        # - add login_redirect_url
                        raise exceptions.Unauthorized(
                            reasons, status_code=status
                        )

                    else:
                        is_authorized = await validate_scopes(
                            request,
                            scopes,
                            user_scopes,
                            require_all=require_all,
                            require_all_actions=require_all_actions,
                            override=override,
                            destructure=destructure,
                            request_args=args,
                            request_kwargs=kwargs,
                        )
                        if not is_authorized:
                            status = 403
                            reasons = "Invalid scope."

                            # TODO:
                            # - add login_redirect_url
                            raise exceptions.Unauthorized(
                                reasons, status_code=status
                            )

            # the user is authorized.
            # run the handler method and return the response
            # NOTE: it's possible to use return await.utils(f, ...) in
            # here, but inside the @protected decorator it wont work,
            # so this is left as is for now
            response = f(request, *args, **kwargs)
            if isawaitable(response):
                response = await response
            return response

        return decorated_function

    return decorator


def inject_user(initialized_on=None, **kw):
    def decorator(f):
        @wraps(f)
        async def decorated_function(request, *args, **kwargs):
            if issubclass(request.__class__, HTTPMethodView):
                request = args[0]

            if initialized_on and isinstance(
                initialized_on, Blueprint
            ):  # noqa
                instance = initialized_on
            else:
                instance = request.app

            with instant_config(instance, request=request, **kw):
                if request.method == "OPTIONS":
                    return await utils.call(
                        f, request, *args, **kwargs
                    )  # noqa

                payload = await instance.auth.extract_payload(
                    request, verify=False
                )
                user = await utils.call(
                    instance.auth.retrieve_user, request, payload
                )
                response = f(request, user=user, *args, **kwargs)
                return await response

        return decorated_function

    return decorator



from sanic.response import json, text
from sanic.views import HTTPMethodView

from . import exceptions, utils
from .base import BaseDerivative
from .decorators import protected


class BaseEndpoint(BaseDerivative, HTTPMethodView):
    def __init__(self, responses, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.responses = responses

    async def options(self, request, *args, **kwargs):
        return text("", status=204)

    async def do_incoming(self, request, args, kwargs):
        return request, args, kwargs

    async def do_output(self, output):
        return output

    async def do_response(self, response):
        return response


class AuthenticateEndpoint(BaseEndpoint):
    async def post(self, request, *args, **kwargs):
        request, args, kwargs = await self.do_incoming(request, args, kwargs)

        config = self.config
        user = await utils.call(
            self.instance.auth.authenticate, request, *args, **kwargs
        )

        access_token, output = await self.responses.get_access_token_output(
            request, user, self.config, self.instance
        )

        if config.refresh_token_enabled():
            refresh_token = await utils.call(
                self.instance.auth.generate_refresh_token, request, user
            )
            output.update({config.refresh_token_name(): refresh_token})
        else:
            refresh_token = None

        output.update(
            self.responses.extend_authenticate(
                request,
                user=user,
                access_token=access_token,
                refresh_token=refresh_token,
            )
        )

        output = await self.do_output(output)

        resp = self.responses.get_token_response(
            request,
            access_token,
            output,
            refresh_token=refresh_token,
            config=self.config,
        )

        return await self.do_response(resp)


class RetrieveUserEndpoint(BaseEndpoint):
    async def get(self, request, *args, **kwargs):
        request, args, kwargs = await self.do_incoming(request, args, kwargs)

        config = self.config
        if not hasattr(self.instance.auth, "retrieve_user"):
            # NOTE: this condition is only true if `retrieve_user` is wipped
            # out of the `Authentication` class, so it won't happen "easily".
            raise exceptions.MeEndpointNotSetup()  # noqa

        payload = await self.instance.auth.extract_payload(request)
        user = await utils.call(
            self.instance.auth.retrieve_user, request, payload
        )

        if not user:  # noqa
            me = None
        else:
            if isinstance(user, dict):
                me = user
            elif hasattr(user, "to_dict"):
                me = await utils.call(user.to_dict)
            elif hasattr(user, "__json__"):
                me = await utils.call(user.__json__)
            else:
                # implementations ago there was an error where "me" was
                # being used before assignment, so this exception is raised.
                # it'll stay here for now, with a noqa flag
                raise exceptions.InvalidRetrieveUserObject  # noqa

        output = {"me": me}

        output.update(
            self.responses.extend_retrieve_user(
                request, user=user, payload=payload
            )
        )

        output = await self.do_output(output)

        resp = json(output)

        if payload is None and config.cookie_set():  # noqa
            key = config.cookie_access_token_name()
            del resp.cookies[key]

        return await self.do_response(resp)


class VerifyEndpoint(BaseEndpoint):
    async def get(self, request, *args, **kwargs):
        request, args, kwargs = await self.do_incoming(request, args, kwargs)

        is_valid, status, reason = await self.instance.auth._verify(
            request, raise_missing=True, *args, **kwargs
        )

        output = {"valid": is_valid}

        if reason:
            if not isinstance(reason, list):  # noqa
                reason = [reason]
            output.update({"reasons": reason})

        if not is_valid:
            output.update({"exception": exceptions.InvalidToken.__name__})

        output.update(self.responses.extend_verify(request))
        output = await self.do_output(output)
        resp = json(output, status=status)

        return await self.do_response(resp)


class RefreshEndpoint(BaseEndpoint):
    async def post(self, request, *args, **kwargs):
        request, args, kwargs = await self.do_incoming(request, args, kwargs)

        # TODO:
        # - Add more exceptions
        payload = await self.instance.auth.extract_payload(
            request, verify=False
        )

        try:
            user = await utils.call(
                self.instance.auth.retrieve_user, request, payload=payload
            )
        except exceptions.MeEndpointNotSetup:
            message = "Refresh tokens have not been enabled properly."
            "Perhaps you forgot to initialize with a retrieve_user handler?"
            raise exceptions.RefreshTokenNotImplemented(message=message)

        user_id = await self.instance.auth._get_user_id(user)
        refresh_token = await utils.call(
            self.instance.auth.retrieve_refresh_token,
            request=request,
            user_id=user_id,
        )
        if isinstance(refresh_token, bytes):
            refresh_token = refresh_token.decode("utf-8")

        token = await self.instance.auth.retrieve_refresh_token_from_request(
            request
        )

        if refresh_token != token:
            raise exceptions.AuthenticationFailed()

        access_token, output = await self.responses.get_access_token_output(
            request, user, self.config, self.instance
        )

        output.update(
            self.responses.extend_refresh(
                request,
                user=user,
                access_token=access_token,
                refresh_token=refresh_token,
                purported_token=token,
                payload=payload,
            )
        )
        output = await self.do_output(output)

        resp = self.responses.get_token_response(
            request, access_token, output, config=self.config
        )

        return await self.do_response(resp)



__version__ = "1.5.0"
__author__ = "Adam Hopkins"
__credits__ = "Richard Kuesters"

import logging

from .authentication import Authentication
from .claim import Claim
from .configuration import Configuration
from .decorators import inject_user, protected, scoped
from .endpoints import BaseEndpoint
from .initialization import Initialize, initialize
from .responses import Responses

logging.getLogger(__name__).addHandler(logging.NullHandler())


__all__ = [
    "Authentication",
    "BaseEndpoint",
    "Claim",
    "Configuration",
    "initialize",
    "Initialize",
    "inject_user",
    "protected",
    "Responses",
    "scoped",
]



import asyncio
import copy
import logging

from . import exceptions, utils
from .cache import get_cached, is_cached, to_cache

defaults = {
    "access_token_name": "access_token",
    "algorithm": "HS256",
    "auth_mode": True,
    "authorization_header": "authorization",
    "authorization_header_prefix": "Bearer",
    "authorization_header_refresh_prefix": "Refresh",
    "claim_aud": None,
    "claim_iat": False,
    "claim_iss": None,
    "claim_nbf": False,
    "claim_nbf_delta": 0,
    "cookie_access_token_name": "access_token",
    "cookie_domain": "",
    "cookie_expires": None,
    "cookie_httponly": True,
    "cookie_max_age": 0,
    "cookie_path": "/",
    "cookie_refresh_token_name": "refresh_token",
    "cookie_secure": False,
    "cookie_set": False,
    "cookie_split": False,
    "cookie_split_signature_name": "access_token_signature",
    "cookie_strict": True,
    "debug": False,
    "do_protection": True,
    "expiration_delta": 60 * 5 * 6,
    "generate_refresh_token": utils.generate_token,
    "leeway": 60 * 3,
    "login_redirect_url": "/index.html",
    "path_to_authenticate": "/",
    "path_to_refresh": "/refresh",
    "path_to_retrieve_user": "/me",
    "path_to_verify": "/verify",
    "private_key": None,
    "query_string_access_token_name": "access_token",
    "query_string_refresh_token_name": "refresh_token",
    "query_string_set": False,
    "query_string_strict": True,
    "refresh_token_enabled": False,
    "refresh_token_name": "refresh_token",
    "scopes_enabled": False,
    "scopes_name": "scopes",
    "secret": "This is a big secret. Shhhhh",
    "strict_slashes": False,
    "user_secret_enabled": False,
    "url_prefix": "/auth",
    "user_id": "user_id",
    "blueprint_name": "auth_bp",
    "verify_exp": True,
}

aliases = {
    "cookie_access_token_name": "cookie_token_name",
    "secret": "public_key",
    "cookie_split": "split_cookie",
}

ignore_keys = (
    "add_scopes_to_payload",
    "authenticate",
    "class_views",
    "extend_payload",
    "retrieve_refresh_token",
    "retrieve_user",
    "store_refresh_token",
    "destructure_scopes",
)

logger = logging.getLogger(__name__)


def _warn_key(key):
    if key not in ignore_keys:
        logger.warning(
            "Configuration key '%s' found is not valid for sanic-jwt", key
        )


def _create_or_overwrite_config_item(value, key, item_aliases, instance):
    setattr(
        instance,
        key,
        ConfigItem(
            value, item_name=key, config=instance, aliases=item_aliases
        ),
    )


def _update_config_item(key, item_aliases, instance):
    getattr(instance, key)._item_name = key
    getattr(instance, key)._aliases = item_aliases
    getattr(instance, key)._config = instance


class ConfigItem:
    def __init__(
        self,
        value,
        item_name=None,
        config=None,
        inject_request=True,
        aliases=None,
    ):
        self._value = value
        self._item_name = item_name
        self._config = config
        self._inject_request = inject_request
        self._override = False
        self._override_value = None

        if aliases is not None and isinstance(aliases, (list, tuple, set)):
            self._aliases = aliases
        else:
            self._aliases = []

    def update(self, value):
        self._value = value

    def __call__(self, **kwargs):
        if self._override:
            return self._override_value

        if asyncio.get_event_loop().is_running():
            if is_cached(self._item_name):
                return get_cached(self._item_name)

            if self._get_from_config is not None:
                args = []

                if self._inject_request and is_cached("_request"):  # noqa
                    args.append(get_cached("_request"))
                val = self._get_from_config.__call__(*args)
                to_cache(self._item_name, val)
                return val

        return self._value

    @property
    def _get_from_config(self):
        if hasattr(self._config, self._get_fn):
            return getattr(self._config, self._get_fn)

        return None

    @property
    def _get_fn(self):
        return "get_{}".format(self._item_name)

    @property
    def aliases(self):
        return self._aliases


class Configuration:
    def __iter__(self):  # noqa
        for key in self.config_keys:
            yield getattr(self, key)

        for key in self.aliases_keys:
            yield getattr(self, key)

    def __contains__(self, item):
        return item in self.all_config_keys

    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)

        _defaults = copy.deepcopy(defaults)
        _aliases = copy.deepcopy(aliases)
        _config_keys = []

        if args and isinstance(args[0], dict):  # noqa
            _args = cls.extract_presets(args[0])
            for key, value in _args.items():
                if key in _defaults or key in _aliases:
                    _defaults.update({key: value})

        for key, value in _defaults.items():
            item_aliases = []
            if key in _aliases:
                item_aliases = [aliases.get(key)]

            # check if a configuration key is set
            # and is an instance of ConfigItem
            if hasattr(instance, key) and isinstance(
                getattr(instance, key), ConfigItem
            ):
                _update_config_item(key, item_aliases, instance)
            # check if a configuration key is set with a value
            elif hasattr(instance, key):
                val = getattr(instance, key)
                _create_or_overwrite_config_item(
                    val, key, item_aliases, instance
                )
            else:
                _create_or_overwrite_config_item(
                    value, key, item_aliases, instance
                )

            # check if a setter is available on config class
            fn_name = "set_{}".format(key)
            if hasattr(instance, fn_name):
                set_fn = getattr(instance, fn_name)
                if not callable(set_fn):
                    logger.warning(
                        'variable "%s" set in Configuration is not callable',
                        fn_name,
                    )
                    continue

                val = set_fn.__call__()
                if isinstance(val, ConfigItem):
                    setattr(instance, key, val)
                    _update_config_item(key, item_aliases, instance)
                else:
                    _create_or_overwrite_config_item(
                        val, key, item_aliases, instance
                    )

            # 'reference' aliases
            for alias in item_aliases:
                setattr(instance, alias, getattr(instance, key))

            _config_keys.append(key)

        setattr(instance, "_config_keys", _config_keys)
        setattr(instance, "_config_aliases", _aliases)
        setattr(instance, "_config_aliases_keys", _aliases.values())
        setattr(
            instance,
            "_all_config_keys",
            _config_keys + list(_aliases.values()),
        )
        setattr(
            instance,
            "_config_alias_mapping",
            {alias: key for key, alias in _aliases.items()},
        )

        return instance

    def get(self, item):
        """Helper method to avoid calling getattr
        """
        if item in self:  # noqa
            item = getattr(self, item)
            return item()

    @property
    def config_keys(self):
        return self._config_keys

    @property
    def config_aliases(self):
        return self._config_aliases

    @property
    def all_config_keys(self):
        return self._all_config_keys

    @property
    def config_aliases_keys(self):
        return self._config_aliases_keys

    @property
    def config_alias_mapping(self):
        return self._config_alias_mapping

    def __init__(self, app_config, **kwargs):
        for key, value in kwargs.items():
            self._merge(key, value)

        self._validate_secret()
        self._validate_keys()
        self._load_keys()
        self._overrides = None

    def _merge(self, key, value):
        if key in self.config_keys:
            item = getattr(self, key)
            item.update(value)
            for alias in item.aliases:
                self._merge(alias, value)
        elif key in self.config_aliases_keys:
            correct_key = None
            for v in self.config_aliases.values():  # noqa
                if key == v:
                    correct_key = key
                    break

            if hasattr(self, correct_key):  # noqa
                getattr(self, correct_key).update(value)
        else:
            _warn_key(key)

    def _validate_secret(self):
        logger.debug("validating provided secret")
        if self.secret() is None or (
            isinstance(self.secret(), str) and self.secret().strip() == ""
        ):
            raise exceptions.InvalidConfiguration(
                "the SANIC_JWT_SECRET parameter cannot be None nor an empty "
                "string"
            )

    def _validate_keys(self):
        logger.debug("validating keys (if needed)")
        if utils.algorithm_is_asymmetric(self.algorithm()) and (
            self.private_key() is None
            or (
                isinstance(self.private_key(), str)
                and self.private_key().strip() == ""
            )
        ):
            raise exceptions.RequiredKeysNotFound

    def _load_keys(self):
        logger.debug("loading secret and/or keys (if needed)")
        try:
            self.secret.update(utils.load_file_or_str(self.secret()))
            if utils.algorithm_is_asymmetric(self.algorithm()):
                self.private_key.update(
                    utils.load_file_or_str(self.private_key())
                )
        except exceptions.ProvidedPathNotFound as exc:
            if utils.algorithm_is_asymmetric(self.algorithm()):
                raise exceptions.RequiredKeysNotFound

            raise exc  # noqa

    def _do_overrides(self, cleanup=False, **kwargs):
        for key, value in kwargs.items():
            if key in self.config_keys:
                item = getattr(self, key)
                item._override = False if cleanup else True
                item._override_value = None if cleanup else value
            elif key in self.config_aliases_keys:
                correct_key = self.config_alias_mapping.get(key)
                item = getattr(self, correct_key)
                item._override = False if cleanup else True
                item._override_value = None if cleanup else value

    @staticmethod
    def extract_presets(app_config):
        """
        Pull the application's configurations for Sanic JWT
        """
        return {
            x.lower()[10:]: app_config.get(x)
            for x in filter(lambda x: x.startswith("SANIC_JWT"), app_config)
        }



from sanic.exceptions import SanicException
from sanic.exceptions import Unauthorized as SanicUnauthorized


class SanicJWTException(SanicException):
    pass


class InvalidToken(SanicJWTException):
    pass


class AuthenticationFailed(SanicJWTException):
    status_code = 401

    def __init__(self, message="Authentication failed.", **kwargs):
        super().__init__(message, **kwargs)


class MissingAuthorizationHeader(SanicJWTException):
    status_code = 400

    def __init__(self, message="Authorization header not present.", **kwargs):
        super().__init__(message, **kwargs)


class MissingAuthorizationCookie(SanicJWTException):
    status_code = 400

    def __init__(self, message="Authorization cookie not present.", **kwargs):
        super().__init__(message, **kwargs)


class MissingAuthorizationQueryArg(SanicJWTException):
    status_code = 400

    def __init__(
        self, message="Authorization query argument not present.", **kwargs
    ):
        super().__init__(message, **kwargs)


class InvalidAuthorizationHeader(SanicJWTException):
    status_code = 400

    def __init__(self, message="Authorization header is invalid.", **kwargs):
        super().__init__(message, **kwargs)


class InvalidCustomClaim(SanicJWTException):
    status_code = 500

    def __init__(self, message="Custom claim is invalid.", **kwargs):
        super().__init__(message, **kwargs)


class InvalidCustomClaimError(SanicJWTException):
    status_code = 401

    def __init__(self, message="Custom claim value was not met.", **kwargs):
        super().__init__(message, **kwargs)


class InvalidVerification(SanicJWTException):
    status_code = 500

    def __init__(
        self,
        message="Verifications must be a callable object "
        "returning a boolean value.",
        **kwargs
    ):
        super().__init__(message, **kwargs)


class InvalidVerificationError(SanicJWTException):
    status_code = 401

    def __init__(self, message="Verifications were not met.", **kwargs):
        super().__init__(message, **kwargs)


class AuthenticateNotImplemented(SanicJWTException):
    status_code = 500

    def __init__(
        self,
        message="Sanic JWT initialized without providing an authenticate "
        "method.",
        **kwargs
    ):
        super().__init__(message, **kwargs)


class RefreshTokenNotImplemented(SanicJWTException):
    status_code = 500

    def __init__(
        self, message="Refresh tokens have not been enabled.", **kwargs
    ):
        super().__init__(message, **kwargs)


class ScopesNotImplemented(SanicJWTException):
    status_code = 500

    def __init__(
        self,
        message="Scopes have not been enabled. Initialize with "
        "add_scopes_to_payload to provide scoping.",
        **kwargs
    ):
        super().__init__(message, **kwargs)


class UserSecretNotImplemented(SanicJWTException):
    status_code = 500

    def __init__(
        self, message="User secrets have not been enabled.", **kwargs
    ):
        super().__init__(message, **kwargs)


class MissingRegisteredClaim(SanicJWTException):
    status_code = 500

    def __init__(
        self,
        message="One or more claims have been registered, but your "
        "extend_payload() method does not supply them. ",
        missing=None,
        **kwargs
    ):
        if missing:  # noqa
            message += str(missing)
        super().__init__(message, **kwargs)


class MeEndpointNotSetup(SanicJWTException):
    status_code = 500

    def __init__(
        self,
        message="/me endpoint has not been setup. Pass retrieve_user if "
        "you with to proceeed.",
        **kwargs
    ):
        super().__init__(message, **kwargs)


class InvalidRetrieveUserObject(SanicJWTException):
    status_code = 500

    def __init__(
        self,
        message="The retrieve_user method should return either a dict or "
        "an object with a to_dict or __json__ method.",
        **kwargs
    ):
        super().__init__(message, **kwargs)


class InitializationFailure(SanicJWTException):
    status_code = 500

    def __init__(
        self,
        message="Sanic JWT was not initialized properly. It must be "
        "instantiated on a sanic.Sanic or sanic.Blueprint "
        "instance.",
        **kwargs
    ):
        super().__init__(message, **kwargs)


class Unauthorized(SanicJWTException, SanicUnauthorized):
    def __init__(self, message="Auth required.", **kwargs):
        super().__init__(message, scheme="Bearer", **kwargs)


class InvalidClassViewsFormat(SanicJWTException):
    def __init__(
        self,
        message="class_views should follow this format ('<SOME ROUTE>', "
        "ClassInheritedFromBaseEndpoint)",
        **kwargs
    ):
        super().__init__(message, **kwargs)


class InvalidConfiguration(SanicJWTException):
    def __init__(self, message="", **kwargs):
        message = (
            "An invalid setting was passed to the Sanic JWT "
            "configuration: " + str(message)
        )
        super().__init__(message, **kwargs)


class InvalidPayload(SanicJWTException):
    status_code = 500

    def __init__(self, message="", **kwargs):
        message = (
            "Payload must be a dictionary with a key mapped to "
            "SANIC_JWT_USER_ID"
        )
        super().__init__(message, **kwargs)


class RequiredKeysNotFound(SanicJWTException):
    def __init__(
        self,
        message="You must provide both (valid) SANIC_JWT_PUBLIC_KEY and "
        "SANIC_JWT_PRIVATE_KEY when using asymmetric "
        "cryptographic algorithms like RS*, EC* or PS*",
        **kwargs
    ):
        super().__init__(message, **kwargs)


class ProvidedPathNotFound(SanicJWTException):
    def __init__(
        self, message="The Path object given is not a valid file", **kwargs
    ):
        super().__init__(message, **kwargs)


class LoopNotRunning(SanicJWTException):
    def __init__(
        self, message="The asyncio loop is not currently running", **kwargs
    ):
        super().__init__(message, **kwargs)



from sanic_jwt import exceptions


class Claim:
    def __init__(self):
        required = ("key", "setup", "verify")
        if any(not hasattr(self, x) for x in required):
            raise exceptions.InvalidCustomClaim()

    @classmethod
    def _register(cls, sanicjwt):
        instance = cls()
        sanicjwt.instance.auth._custom_claims.add(instance)

    def get_key(self):
        return self.key

    def _verify(self, payload):
        key = self.get_key()
        value = payload.get(key)
        valid_claim = self.verify(value)
        if not isinstance(valid_claim, bool):
            raise exceptions.InvalidCustomClaim()

        if valid_claim is False:
            message = "Invalid claim: {}".format(key)
            raise exceptions.InvalidCustomClaimError(message=message)



import inspect
from collections import namedtuple

from sanic import Blueprint, Sanic

from sanic_jwt import endpoints, exceptions
from sanic_jwt.authentication import Authentication
from sanic_jwt.configuration import Configuration
from sanic_jwt.decorators import inject_user, protected, scoped
from sanic_jwt.responses import Responses

_Handler = namedtuple(
    "_Handler", ["name", "keys", "exception", "outside_auth_mode"]
)
_EndpointMapping = namedtuple(
    "_EndpointMapping",
    ["cls", "endpoint", "keys", "is_protected", "protected_kwargs"],
)


def initialize(*args, **kwargs):
    """
    Functional approach to initializing Sanic JWT. This was the original
    method, but was replaced by the Initialize class. It is recommended to use
    the class because it is more flexible. There is no current plan to remove
    this method, but it may be depracated in the future.
    """
    if len(args) > 1:
        kwargs.update({"authenticate": args[1]})
    return Initialize(args[0], **kwargs)


endpoint_mappings = (
    _EndpointMapping(
        endpoints.AuthenticateEndpoint,
        "authenticate",
        ["auth_mode"],
        False,
        {},
    ),
    _EndpointMapping(
        endpoints.RetrieveUserEndpoint,
        "retrieve_user",
        ["auth_mode"],
        True,
        {},
    ),
    _EndpointMapping(
        endpoints.VerifyEndpoint, "verify", ["auth_mode"], False, {}
    ),
    _EndpointMapping(
        endpoints.RefreshEndpoint,
        "refresh",
        ["auth_mode", "refresh_token_enabled"],
        True,
        {"verify_exp": False},
    ),
)

auth_mode_handlers = (
    _Handler(
        "authenticate", None, exceptions.AuthenticateNotImplemented(), False
    ),
    _Handler(
        "store_refresh_token",
        ["refresh_token_enabled"],
        exceptions.RefreshTokenNotImplemented(),
        False,
    ),
    _Handler(
        "retrieve_refresh_token",
        ["refresh_token_enabled"],
        exceptions.RefreshTokenNotImplemented(),
        False,
    ),
    _Handler(
        "add_scopes_to_payload",
        ["scopes_enabled"],
        exceptions.ScopesNotImplemented(),
        False,
    ),
    _Handler(
        "retrieve_user_secret",
        ["user_secret_enabled"],
        exceptions.UserSecretNotImplemented(),
        False,
    ),
    _Handler("override_scope_validator", None, None, False),
    _Handler("destructure_scopes", None, None, False),
    _Handler("extend_payload", None, None, False),
)
auth_mode_agnostic_handlers = (_Handler("retrieve_user", None, None, True),)
handlers = auth_mode_handlers + auth_mode_agnostic_handlers

init_classes = (
    "configuration_class",
    "authentication_class",
    "responses_class",
)


class Initialize:
    """Class used to initialize Sanic JWT

    Must be initialized with a keyword argument: `authenticate` that is a
    method that MUST return a user object that when iterated upon returns a
    dictionary, or has a `to_dict` method. The resulting dictionary MUST
    have a key/value for a unique user id.
    """

    authentication_class = Authentication
    configuration_class = Configuration
    responses_class = Responses

    def __init__(self, instance, app=None, **kwargs):
        for class_name in init_classes:
            if class_name in kwargs:
                value = kwargs.pop(class_name)
                setattr(self, class_name, value)

        app = self.__get_app(instance, app=app)

        self.app = app
        self.kwargs = kwargs
        self.instance = instance
        self.config = None
        self.bp = None

        self.__check_deprecated()
        self.__check_classes()
        self.__load_configuration()
        self.__initialize_bp()
        self.__load_responses()
        self.__add_class_views()
        self.__add_endpoints()
        self.__initialize_instance()
        self.__initialize_claims()

    def __check_deprecated(self):
        """
        Checks for deprecated configuration keys
        """
        # Depracation notices
        if "SANIC_JWT_HANDLER_PAYLOAD_SCOPES" in self.app.config:
            raise exceptions.InvalidConfiguration(
                "SANIC_JWT_HANDLER_PAYLOAD_SCOPES has been deprecated. "
                "Instead, pass your handler method (not an import path) as "
                "initialize(add_scopes_to_payload=my_scope_extender)"
            )

        if "SANIC_JWT_PAYLOAD_HANDLER" in self.app.config:
            raise exceptions.InvalidConfiguration(
                "SANIC_JWT_PAYLOAD_HANDLER has been deprecated. "
                "Instead, you will need to subclass Authentication. "
            )

        if "SANIC_JWT_HANDLER_PAYLOAD_EXTEND" in self.app.config:
            raise exceptions.InvalidConfiguration(
                "SANIC_JWT_HANDLER_PAYLOAD_EXTEND has been deprecated. "
                "Instead, you will need to subclass Authentication. "
                "Check out the documentation for more information."
            )

    def __add_endpoints(self):
        """
        Initialize the Sanic JWT Blueprint and add to the instance initialized
        """
        for mapping in endpoint_mappings:
            if all(map(self.config.get, mapping.keys)):
                self.__add_single_endpoint(
                    mapping.cls,
                    mapping.endpoint,
                    mapping.is_protected,
                    mapping.protected_kwargs,
                )

        self.bp.exception(exceptions.SanicJWTException)(
            self.responses.exception_response
        )

        if not self.instance_is_blueprint:
            url_prefix = self._get_url_prefix()
            self.instance.blueprint(self.bp, url_prefix=url_prefix)

    def __add_class_views(self):
        """
        Include any custom class views on the Sanic JWT Blueprint
        """
        config = self.config
        if "class_views" in self.kwargs:
            class_views = self.kwargs.pop("class_views")

            for route, view in class_views:
                if issubclass(view, endpoints.BaseEndpoint) and isinstance(
                    route, str
                ):
                    self.bp.add_route(
                        view.as_view(
                            self.responses,
                            config=self.config,
                            instance=self.instance,
                        ),
                        route,
                        strict_slashes=config.strict_slashes(),
                    )
                else:
                    raise exceptions.InvalidClassViewsFormat()

    def __check_classes(self):
        """
        Check if any of the default classes (`Authentication`, `Configuration`
        and / or `Responses`) have been overwitten and if they're still valid
        """
        # msg took from BaseAuthentication
        msg = (
            "Sanic JWT was not initialized properly. It did not received "
            "an instance of {}"
        )
        if not issubclass(self.authentication_class, Authentication):
            raise exceptions.InitializationFailure(
                message=msg.format("Authentication")
            )

        if not issubclass(self.configuration_class, Configuration):
            raise exceptions.InitializationFailure(
                message=msg.format("Configuration")
            )

        if not issubclass(self.responses_class, Responses):
            raise exceptions.InitializationFailure(
                message=msg.format("Responses")
            )

    def __initialize_instance(self):
        """
        Take any predefined methods/handlers and insert them into Sanic JWT
        """
        config = self.config

        # Initialize instance of the Authentication class
        self.instance.auth = self.authentication_class(self.app, config=config)

        init_handlers = (
            handlers if config.auth_mode() else auth_mode_agnostic_handlers
        )

        for handler in init_handlers:
            if handler.keys is None:
                self.__check_method_in_auth(handler.name, handler.exception)
            else:
                if all(map(config.get, handler.keys)):
                    self.__check_method_in_auth(
                        handler.name, handler.exception
                    )

        for handler in init_handlers:
            if handler.name in self.kwargs:
                method = self.kwargs.pop(handler.name)
                setattr(self.instance.auth, handler.name, method)

    def __initialize_claims(self):
        if "extra_verifications" in self.kwargs:
            self.instance.auth._extra_verifications = self.kwargs.get(
                "extra_verifications"
            )

        if "custom_claims" in self.kwargs:
            try:
                for claim in self.kwargs.get("custom_claims"):
                    claim._register(self)
            except AttributeError:
                raise exceptions.InvalidCustomClaim

    def __check_method_in_auth(self, method_name, exc):
        if method_name not in self.kwargs:
            method_impl = getattr(self.instance.auth, method_name)
            if not inspect.ismethod(method_impl):
                self.__raise_if_not_none(exc)
            if method_impl.__func__ == getattr(Authentication, method_name):
                self.__raise_if_not_none(exc)

            self.kwargs.update({method_name: method_impl})

    def __load_configuration(self):
        """
        Configure settings for the instance in the following order:

        1. Sanic JWT's defaults
        2. Custom Configuration class
        3. Key word arguments passed to Initialize
        """
        handler_to_enable = filter(lambda h: h.keys is not None, handlers)
        for handler in handler_to_enable:
            if handler.name in self.kwargs:
                for k in handler.keys:
                    self.kwargs.update({k: True})

        self.config = self.configuration_class(self.app.config, **self.kwargs)

    def __load_responses(self):
        self.responses = self.responses_class(self.config, self.instance)

    def __add_single_endpoint(
        self, endpoint_cls, path_name, is_protected, protected_kwargs
    ):
        path_name = getattr(self.config, "path_to_{}".format(path_name))()
        if is_protected:
            endpoint_cls.decorators = [self.protected(**protected_kwargs)]
        if self.instance_is_blueprint:
            path_name = self._get_url_prefix() + path_name
            if self.instance.url_prefix:
                path_name = path_name.replace(self.instance.url_prefix, "")
            self.instance.add_route(
                endpoint_cls.as_view(
                    config=self.config,
                    instance=self.instance,
                    responses=self.responses,
                ),
                path_name,
            )
        else:
            self.bp.add_route(
                endpoint_cls.as_view(
                    config=self.config,
                    instance=self.instance,
                    responses=self.responses,
                ),
                path_name,
            )

    def _get_url_prefix(self):
        bp_url_prefix = (
            self.bp.url_prefix if self.bp.url_prefix is not None else ""
        )
        config_url_prefix = self.config.url_prefix()
        url_prefix = bp_url_prefix + config_url_prefix
        return url_prefix

    @staticmethod
    def __get_app(instance, app=None):
        if isinstance(instance, Sanic):
            return instance

        elif isinstance(instance, Blueprint):
            if app is not None:
                return app
            message = "Sanic JWT was not initialized properly. When "
            "instantiating on a sanic.Blueprint you must pass your app "
            "instance: Initialize(bp, app=app, ...)"
            raise exceptions.InitializationFailure(message=message)

        raise exceptions.InitializationFailure

    def __initialize_bp(self):
        if isinstance(self.instance, Sanic):
            bp_name = self.config.blueprint_name()
            self.bp = Blueprint(bp_name)

        elif isinstance(self.instance, Blueprint):
            self.bp = self.instance

        else:
            raise exceptions.InitializationFailure  # noqa see line above

    def protected(self, *args, **kwargs):
        args = list(args)
        args.insert(0, self.instance)
        return protected(*args, **kwargs)

    def scoped(self, scopes, **kwargs):
        kwargs.update({"initialized_on": self.instance})
        return scoped(scopes, **kwargs)

    def inject_user(self, *args, **kwargs):
        args = list(args)
        args.insert(0, self.instance)
        return inject_user(*args, **kwargs)

    @property
    def instance_is_blueprint(self):
        return isinstance(self.instance, Blueprint)

    @staticmethod
    def __raise_if_not_none(exc):
        if exc is not None:
            raise exc



import codecs
import re
from os import path

from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))


def open_local(paths, mode="r", encoding="utf8"):
    p = path.join(here, *paths)
    return codecs.open(p, mode, encoding)


with open_local(["sanic_jwt", "__init__.py"], encoding="latin1") as fp:
    try:
        version = re.findall(
            r"^__version__ = \"([0-9\.]+)\"", fp.read(), re.M
        )[0]
    except IndexError:
        raise RuntimeError("Unable to determine version.")

with open_local(["README.md"]) as rm:
    long_description = rm.read()

extras_require = {"docs": ["Sphinx"]}

extras_require["all"] = []
for reqs in extras_require.values():
    extras_require["all"].extend(reqs)

install_requires = ["pyjwt"]

setup(
    name="sanic-jwt",
    version=version,
    description="JWT oauth flow for Sanic",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ahopkins/sanic-jwt",
    download_url="https://github.com/ahopkins/sanic-jwt/archive/master.zip",
    author="Adam Hopkins",
    author_email="admhpkns@gmail.com",
    license="MIT",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    keywords="sanic oauth authentication jwt",
    packages=find_packages(exclude=["example", "tests"]),
    install_requires=install_requires,
    extras_require=extras_require,
    package_data={},
)



import inspect
import logging
import warnings
from contextlib import contextmanager
from datetime import datetime, timedelta

import jwt

from . import exceptions, utils
from .exceptions import (
    InvalidCustomClaimError,
    InvalidVerification,
    InvalidVerificationError,
    SanicJWTException
)

logger = logging.getLogger(__name__)
claim_label = {"iss": "issuer", "iat": "iat", "nbf": "nbf", "aud": "audience"}


class BaseAuthentication:
    def __init__(self, app, config):
        self.app = app
        self.claims = ["exp"]
        self._extra_verifications = None
        self.config = config
        self._reasons = []
        self._custom_claims = set()

    async def _get_user_id(self, user, *, asdict=False):
        """
        Get a user_id from a user object. If `asdict` is True, will return
        it as a dict with `config.user_id` as key. The `asdict` keyword
        defaults to `False`.
        """
        uid = self.config.user_id()
        if isinstance(user, dict):
            user_id = user.get(uid)
        elif hasattr(user, "to_dict"):
            _to_dict = await utils.call(user.to_dict)
            user_id = _to_dict.get(uid)
        else:
            raise exceptions.InvalidRetrieveUserObject()

        if asdict:
            return {uid: user_id}

        return user_id

    async def build_payload(self, user, *args, **kwargs):
        return await self._get_user_id(user, asdict=True)

    async def add_claims(
        self, payload, user, inline_claims=None, *args, **kwargs
    ):
        """
        Injects standard claims into the payload for: exp, iss, iat, nbf, aud.
        And, custom claims, if they exist
        """
        delta = timedelta(seconds=self.config.expiration_delta())
        exp = datetime.utcnow() + delta
        additional = {"exp": exp}

        for option in ["iss", "iat", "nbf", "aud"]:
            setting = "claim_{}".format(option.lower())
            if setting in self.config:  # noqa
                attr = self.config.get(setting)
                if attr:
                    self.claims.append(option)
                    method_name = "build_claim_{}".format(option)
                    method = getattr(utils, method_name)
                    additional.update({option: method(attr, self.config)})

        payload.update(additional)

        if self._custom_claims:
            custom_claims = {}
            for claim in self._custom_claims:
                custom_claims[claim.get_key()] = await utils.call(
                    claim.setup, payload, user
                )
            payload.update(custom_claims)

        if inline_claims:
            claims = {}
            for claim in inline_claims:
                instance = claim()
                claims[instance.get_key()] = await utils.call(
                    instance.setup, payload, user
                )
            payload.update(claims)

        return payload

    async def extend_payload(self, payload, user=None, *args, **kwargs):
        return payload

    async def store_refresh_token(self, *args, **kwargs):
        raise exceptions.RefreshTokenNotImplemented()  # noqa

    async def retrieve_refresh_token(self, *args, **kwargs):
        raise exceptions.RefreshTokenNotImplemented()  # noqa

    async def authenticate(self, *args, **kwargs):
        raise exceptions.AuthenticateNotImplemented()  # noqa

    async def add_scopes_to_payload(self, *args, **kwargs):
        raise exceptions.ScopesNotImplemented()  # noqa

    async def retrieve_user_secret(self, *args, **kwargs):
        raise exceptions.UserSecretNotImplemented()  # noqa

    def override_scope_validator(  # noqa
        self,
        is_valid,
        required,
        user_scopes,
        require_all_actions,
        *args,
        **kwargs,
    ):
        return is_valid

    def destructure_scopes(self, scopes, *args, **kwargs):  # noqa
        return scopes

    async def retrieve_user(self, *args, **kwargs):
        raise exceptions.MeEndpointNotSetup  # noqa


class Authentication(BaseAuthentication):
    async def _check_authentication(
        self, request, request_args, request_kwargs
    ):
        """
        Checks a request object to determine if that request contains a valid,
        and authenticated JWT.

        It returns a tuple:
        1. Boolean whether the request is authenticated with a valid JWT
        2. HTTP status code
        3. Reasons (if any) for a potential authentication failure
        """
        try:
            is_valid, status, reasons = await self._verify(
                request,
                request_args=request_args,
                request_kwargs=request_kwargs,
            )
        except Exception as e:
            logger.debug(e.args)
            if self.config.debug():
                raise e

            args = e.args if isinstance(e, SanicJWTException) else []

            raise exceptions.Unauthorized(*args)

        return is_valid, status, reasons

    async def _decode(self, token, verify=True, inline_claims=None):
        """
        Take a JWT and return a decoded payload. Optionally, will verify
        the claims on the token.
        """
        secret = await self._get_secret(token=token)
        algorithm = self._get_algorithm()
        kwargs = {}

        for claim in self.claims:
            if claim != "exp":
                setting = "claim_{}".format(claim.lower())
                if setting in self.config:  # noqa
                    value = self.config.get(setting)
                    kwargs.update({claim_label[claim]: value})

        kwargs["leeway"] = int(self.config.leeway())
        if "claim_aud" in self.config:  # noqa
            kwargs["audience"] = self.config.claim_aud()
        if "claim_iss" in self.config:  # noqa
            kwargs["issuer"] = self.config.claim_iss()

        decoded = jwt.decode(
            token,
            secret,
            algorithms=[algorithm],
            verify=verify,
            options={"verify_exp": self.config.verify_exp()},
            **kwargs,
        )

        if verify:
            if self._extra_verifications:
                self._verify_extras(decoded)
            if self._custom_claims or inline_claims:
                self._verify_custom_claims(
                    decoded, inline_claims=inline_claims
                )

        return decoded

    def _get_algorithm(self):
        return self.config.algorithm()

    async def _get_payload(self, user, inline_claims=None):
        """
        Given a user object, create a payload and extend it as configured.
        """
        payload = await utils.call(self.build_payload, user)

        if (
            not isinstance(payload, dict)
            or self.config.user_id() not in payload
        ):
            raise exceptions.InvalidPayload

        payload = await utils.call(
            self.add_claims, payload, user, inline_claims
        )

        extend_payload_args = inspect.getfullargspec(self.extend_payload)
        args = [payload]
        if "user" in extend_payload_args.args:
            args.append(user)
        payload = await utils.call(self.extend_payload, *args)

        if self.config.scopes_enabled():
            scopes = await utils.call(self.add_scopes_to_payload, user)
            if not isinstance(scopes, (tuple, list)):
                scopes = [scopes]
            payload[self.config.scopes_name()] = scopes

        claims = self.claims + [x.get_key() for x in self._custom_claims]
        missing = [x for x in claims if x not in payload]
        if missing:
            logger.debug("")
            raise exceptions.MissingRegisteredClaim(missing=missing)

        return payload

    async def _get_refresh_token(self, request):
        """
        Extract a refresh token from a request object.
        """
        return self._get_token(request, refresh_token=True)

    async def _get_secret(self, token=None, payload=None, encode=False):
        if not hasattr(self, "_is_asymmetric"):
            self._is_asymmetric = utils.algorithm_is_asymmetric(
                self._get_algorithm()
            )

        if self.config.user_secret_enabled():
            if not payload:
                algorithm = self._get_algorithm()
                payload = jwt.decode(
                    token, verify=False, algorithms=[algorithm]
                )
            user_id = payload.get("user_id")
            return await utils.call(
                self.retrieve_user_secret,
                user_id=user_id,
                encode=self._is_asymmetric and encode,
            )

        if self._is_asymmetric and encode:
            return self.config.private_key()

        return self.config.secret()

    def _get_token_from_cookies(self, request, refresh_token):
        """
        Extract the token if present inside the request cookies.
        """
        if refresh_token:
            cookie_token_name_key = "cookie_refresh_token_name"
        else:
            cookie_token_name_key = "cookie_access_token_name"
        cookie_token_name = getattr(self.config, cookie_token_name_key)
        token = request.cookies.get(cookie_token_name(), None)
        if not refresh_token and self.config.cookie_split() and token:
            signature_name = self.config.cookie_split_signature_name()
            token += "." + request.cookies.get(signature_name, "")
        return token

    def _get_token_from_headers(self, request, refresh_token):
        """
        Extract the token if present inside the headers of a request.
        """
        header = request.headers.get(self.config.authorization_header(), None)

        if header is None:
            return None

        else:
            header_prefix_key = "authorization_header_prefix"
            header_prefix = getattr(self.config, header_prefix_key)
            if header_prefix():
                try:
                    prefix, token = header.split(" ")
                    if prefix != header_prefix():
                        raise Exception

                except Exception:
                    raise exceptions.InvalidAuthorizationHeader()

            else:
                token = header

            if refresh_token:
                token = request.json.get(self.config.refresh_token_name())

            return token

    def _get_token_from_query_string(self, request, refresh_token):
        """
        Extract the token if present from the request args.
        """
        if refresh_token:
            query_string_token_name_key = "query_string_refresh_token_name"
        else:
            query_string_token_name_key = "query_string_access_token_name"
        query_string_token_name = getattr(
            self.config, query_string_token_name_key
        )
        return request.args.get(query_string_token_name(), None)

    def _get_token(self, request, refresh_token=False):
        """
        Extract a token from a request object.
        """
        if self.config.cookie_set():
            token = self._get_token_from_cookies(request, refresh_token)
            if token:
                return token

            else:
                if self.config.cookie_strict():
                    raise exceptions.MissingAuthorizationCookie()

        if self.config.query_string_set():
            token = self._get_token_from_query_string(request, refresh_token)
            if token:
                return token

            else:
                if self.config.query_string_strict():
                    raise exceptions.MissingAuthorizationQueryArg()

        token = self._get_token_from_headers(request, refresh_token)

        if token:
            return token

        raise exceptions.MissingAuthorizationHeader()

    async def _verify(
        self,
        request,
        return_payload=False,
        verify=True,
        raise_missing=False,
        request_args=None,
        request_kwargs=None,
        *args,
        **kwargs,
    ):
        """
        Verify that a request object is authenticated.
        """
        try:
            token = self._get_token(request)
            is_valid = True
            reason = None
        except (
            exceptions.MissingAuthorizationCookie,
            exceptions.MissingAuthorizationQueryArg,
            exceptions.MissingAuthorizationHeader,
        ) as e:
            token = None
            is_valid = False
            reason = list(e.args)
            status = e.status_code if self.config.debug() else 401

            if raise_missing:
                if not self.config.debug():
                    e.status_code = 401
                raise e

        if token:
            try:
                payload = await self._decode(token, verify=verify)
            except (
                jwt.exceptions.ExpiredSignatureError,
                jwt.exceptions.InvalidIssuerError,
                jwt.exceptions.ImmatureSignatureError,
                jwt.exceptions.InvalidIssuedAtError,
                jwt.exceptions.InvalidAudienceError,
                InvalidVerificationError,
                InvalidCustomClaimError,
            ) as e:
                # Make sure that the reasons all end with '.' for consistency
                reason = [
                    x if x.endswith(".") else "{}.".format(x)
                    for x in list(e.args)
                ]
                payload = None
                status = 401
                is_valid = False
            except jwt.exceptions.DecodeError as e:
                self._reasons = e.args
                # Make sure that the reasons all end with '.' for consistency
                reason = (
                    [
                        x if x.endswith(".") else "{}.".format(x)
                        for x in list(e.args)
                    ]
                    if self.config.debug()
                    else "Auth required."
                )
                logger.debug(e.args)
                is_valid = False
                payload = None
                status = 400 if self.config.debug() else 401
        else:
            payload = None

        if return_payload:
            return payload

        status = 200 if is_valid else status

        return is_valid, status, reason

    def _verify_extras(self, payload):
        for verification in self._extra_verifications:
            if not callable(verification):
                raise InvalidVerification()

            verified = verification(payload)
            if not isinstance(verified, bool):
                raise InvalidVerification()

            if verified is False:
                raise InvalidVerificationError()

    def _verify_custom_claims(self, payload, inline_claims=None):
        if self._custom_claims:
            for claim in self._custom_claims:
                claim._verify(payload)
        if inline_claims:
            for claim in inline_claims:
                claim()._verify(payload)

    async def extract_payload(self, request, verify=True, *args, **kwargs):
        """
        Extract a payload from a request object.
        """
        payload = await self._verify(
            request, return_payload=True, verify=verify, *args, **kwargs
        )
        return payload

    async def extract_scopes(self, request):
        """
        Extract scopes from a request object.
        """
        payload = await self.extract_payload(request)
        if not payload:
            return None

        scopes_attribute = self.config.scopes_name()
        return payload.get(scopes_attribute, None)

    async def extract_user_id(self, request):
        """
        Extract a user id from a request object.
        """
        payload = await self.extract_payload(request)
        user_id_attribute = self.config.user_id()
        return payload.get(user_id_attribute, None)

    async def generate_access_token(
        self, user, extend_payload=None, custom_claims=None
    ):
        """
        Generate an access token for a given user.
        """
        payload = await self._get_payload(user, inline_claims=custom_claims)
        secret = await self._get_secret(payload=payload, encode=True)
        algorithm = self._get_algorithm()

        if extend_payload:
            payload = await utils.call(
                extend_payload, payload=payload, user=user
            )

        return jwt.encode(payload, secret, algorithm=algorithm).decode("utf-8")

    async def generate_refresh_token(self, request, user):
        """
        Generate a refresh token for a given user.
        """
        refresh_token = await utils.call(self.config.generate_refresh_token())
        user_id = await self._get_user_id(user)
        await utils.call(
            self.store_refresh_token,
            user_id=user_id,
            refresh_token=refresh_token,
            request=request,
        )
        return refresh_token

    async def is_authenticated(self, request):
        """
        Checks a request object to determine if that request contains a valid,
        and authenticated JWT.
        """
        is_valid, *_ = await self._verify(request)

        return is_valid

    async def retrieve_refresh_token_from_request(self, request):
        return await self._get_refresh_token(request)

    async def verify_token(
        self, token, return_payload=False, custom_claims=None
    ):
        """
        Perform an inline verification of a token.
        """

        payload = await self._decode(token, inline_claims=custom_claims)
        return payload if return_payload else bool(payload)

    @contextmanager
    def override(self, **kwargs):
        self.config._do_overrides(**kwargs)
        yield
        self.config._do_overrides(cleanup=True, **kwargs)
